/* nomnom */

/*  nspirelink
 *  Copyright (C) 2012 Jim Bauwens
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "ticables.h"
#include "tifiles.h"
#include "ticalcs.h"

#include "lua.h"
#include "lauxlib.h"
#include "lualib.h"

#define TRYF(x) { int aaa_; if((aaa_ = (x))) return aaa_; }

#define MAXBUF 1024
#define KEYSLEEP 5000

//Structs to contain device info

CableModel cable_model = CABLE_USB;
CableHandle* cable;

CalcModel calc_model = CALC_NSPIRE;
int port_number = 1;

CalcHandle* calc;



static void print_usage(void) {
    fprintf(stderr, "usage: %s lua_script\n", "main");
    fflush(stderr);
}

static int check_args(int arg_count, char** arg_strings) {
    if (arg_count != 2) {
        print_usage();
        return(0);
    } else {
        return(1);
    }
}

static void err_message(const char* msg) {
    fprintf(stderr, "%s\n", msg);
    fflush(stderr);
}


static const uint8_t LINKGOTDATA[] = { 124, 256-107, 0 };
static const uint8_t SIN_KEY_BYTES[] = { 0, 116, 0 };

static int mod_table[] = {0, 2, 1};

void cmd_s_keypress_event();

void sendACK(){
	cmd_s_keypress_event(calc, LINKGOTDATA);
	g_usleep(KEYSLEEP);
	return;
}

void sendChar(int ascii){
	uint8_t keydata[] = {ascii, 0, 0};
	cmd_s_keypress_event(calc, keydata);
	g_usleep(KEYSLEEP);	
}

//Lua: nspire_send_char(123)
static int nspire_send_char(lua_State *L){
	int ascii = luaL_checknumber(L, 1);
	sendChar(ascii);
	return 0;
}

//Lua: nspire_send_key(123, 123, 123)
static int nspire_send_key(lua_State *L){
	int a = luaL_checknumber(L, 1);
	int b = luaL_checknumber(L, 2);
	int c = luaL_checknumber(L, 3);
	
	uint8_t keydata[] = {a, b, c};
	cmd_s_keypress_event(calc, keydata);
	g_usleep(KEYSLEEP);	
	
	return 0;
}



void sendData(const char *data, int input_length) {
	//SIN is the identifier
	cmd_s_keypress_event(calc, SIN_KEY_BYTES);
	g_usleep(KEYSLEEP);
	
	int i, j;
    for (i = 0, j = 0; i < input_length;) {
        uint32_t octet_a = i < input_length ? data[i++] : 0;
        uint32_t octet_b = i < input_length ? data[i++] : 0;
        uint32_t octet_c = i < input_length ? data[i++] : 0;

        uint32_t triple = (octet_a << 0x10) + (octet_b << 0x08) + octet_c;


		int bl;
		for (bl=3; bl>=0; bl--)
			sendChar(192 + ( (triple >> bl * 6) & 0x3F) );

	}

    for (i = 0; i < mod_table[input_length % 3]; i++)
        sendChar(190);
        
    //End package with SIN
	cmd_s_keypress_event(calc, SIN_KEY_BYTES);  
	g_usleep(KEYSLEEP);  
	
	return;
}

//Lua: nspire_send("string")
static int nspire_send(lua_State* L) {
	const char *data = luaL_checkstring(L, 1);

	int l = strlen(data);

	if (l > 0){
		sendData(data, l);
	} else {
		lua_pushstring(L, "nspire_send expects a string with length > 0");
		lua_error(L);
	}
	
    return 0;
}

uint8_t* screen_16bitcolor_convert(uint8_t* bitmap)
{
	int w = 320;
	int h = 240;
	guchar *bytemap = g_malloc(3 * w * h);
	int i, j;

	for (i = 0; i < h; i++)
	{
		for (j = 0; j < w; j++)
		{
			uint16_t data = ((uint16_t *)bitmap)[w * i + j];

			int pos = w*i + j;

			// Current code: R5 G6 B5. Let's see if it's correct.
			bytemap[3 * pos + 0] = ((data & 0xF800) >> 11) << 3;
			bytemap[3 * pos + 1] = ((data & 0x07E0) >>  5) << 2;
			bytemap[3 * pos + 2] = ((data & 0x001F) >>  0) << 3;
		}
	}

	return bytemap;
}

char sendbuffer[MAXBUF];


void decode(int r, int g, int b, int p){
	int c = 0;
	c += (r >> 3) << 11;
	c += (g >> 2) << 5 ;
	c +=  b >> 3;  
	char aa = c >> 8;
	char bb = c & 0xFF; 
	sendbuffer[p] = aa;
	sendbuffer[p+1] = bb;
	//printf("%c%c",aa, bb);
}


int decodeString(guchar *bytemap){
	int pos;
	int chr = 0;
	for (pos=0; pos<320; pos++){
		int r = bytemap[3*pos+0];
		int g = bytemap[3*pos+1];
		int b = bytemap[3*pos+2];
		if (r == 248 && g == 252 && b == 248) break;
		//printf("Pixel at %d has color rgb(%d,%d,%d)\r\n", pos, r, g, b);
		decode(r,g,b, chr);
		chr+=2;
	}
	
	if (chr>0){
		sendACK();
		//printf("\r\n");
		//sendbuffer[chr++] = '\r';
		//sendbuffer[chr++] = '\n';
		sendbuffer[chr++] = '\0';
		return 1;
	}
	
	return 0;
}



int parseScreen()
{
	CalcScreenCoord sc = { SCREEN_CLIPPED, 0, 0, 0, 0 };
	uint8_t* bitmap = NULL;

	TRYF(ticalcs_calc_recv_screen(calc, &sc, &bitmap));
	guchar *bytemap;
	
	bytemap = screen_16bitcolor_convert(bitmap);
	
	int d = decodeString(bytemap);
	
	g_free(bytemap);
	free(bitmap);
	
	return d;
}

//Lua: nspire_get()
static int nspire_get(lua_State* luaVM) {
	if (parseScreen() == 1){
		lua_pushstring(luaVM, sendbuffer);
		return 1;
	}

    return 0;
}

//Lua: msleep(1000) sleeps 1s
static int msleep_c(lua_State *L){
	long msecs = lua_tointeger(L, -1);
	g_usleep(1000*msecs);
	return 0;
}

//Lua: nspire_send_file(filepath, foldername)
static int nspire_send_file(lua_State *L){
	const char *filename   = luaL_checkstring(L, 1);
	const char *foldername = luaL_checkstring(L, 2);
	
	if (strlen(filename)>0 && strlen(foldername) > 0){
		FileContent* content = tifiles_content_create_regular(calc_model);
		tifiles_file_read_regular(filename, content);
		VarEntry *entry = content->entries[0];
		strcpy(entry->folder, foldername);  

		tifiles_file_display_regular(content);
		ticalcs_calc_send_var(calc, MODE_NORMAL | MODE_SEND_LAST_VAR, content);
		tifiles_content_delete_regular(content);	
	}else{
		lua_pushstring(L, "nspire_send_file: wrong parameters");
		lua_error(L);	
	}
	
	return 0;
}


static void NOLOG () {};



int main(int argc, char** argv)
{
	int err;

	// Init libticables and libticalcs; Handle error/info messages
	{
		g_setenv("G_MESSAGES_DEBUG", "all", FALSE);
		
		ticables_library_init();
		ticalcs_library_init();
		
		g_log_set_handler("ticalcs" , G_LOG_LEVEL_INFO, NOLOG, NULL);
		g_log_set_handler("ticables", G_LOG_LEVEL_INFO, NOLOG, NULL);
	}

	//Check if there is actually a TI-Nspire connected
	{
		int *pids, npids;
		ticables_get_usb_devices(&pids, &npids);

		if (npids<1 && pids[0] != PID_NSPIRE){
			fprintf(stderr, "No TI-Nspire found !\r\n");
			return 1;
		}

		free(pids);
		
		cable = ticables_handle_new(cable_model, port_number);
		if(cable == NULL) return -1;
		if(ticalcs_probe(cable_model, port_number, &calc_model, 1)){
			fprintf(stderr, "No TI-Nspire found !\r\n");
			return 1;
		}
		
		calc = ticalcs_handle_new(calc_model);
		if(calc == NULL) return -1;
		
		err = ticalcs_cable_attach(calc, cable);
	}


	//Ready ? Let's go !
	ticalcs_calc_isready(calc);

	
	/*
	//////////////

	/////////////

	sendData(calc, buffer, strlen(buffer));
	d = parseScreen(calc);
*/


	if (!check_args(argc, argv)) return(EXIT_FAILURE);
	
	/* Initialize Lua */
	lua_State* luaVM = luaL_newstate();
	if (luaVM == NULL) {
		err_message("Cannot initialize Lua; exiting...");
		return EXIT_FAILURE;
	}
	
	
	/* Stop the GC during Lua library (all std) initialization and function
	* registrations. C functions that are callable from a Lua script must be
	* registered with Lua.
	*/
	
	lua_gc(luaVM, LUA_GCSTOP, 0);
	luaL_openlibs(luaVM);
	lua_register(luaVM, "nspire_send", nspire_send);
	lua_register(luaVM, "nspire_get" , nspire_get);
	lua_register(luaVM, "nspire_send_file", nspire_send_file);
	lua_register(luaVM, "msleep"     , msleep_c);
	lua_register(luaVM, "nspire_send_char" ,nspire_send_char);
	lua_register(luaVM, "nspire_send_key"  ,nspire_send_key );
	
	lua_gc(luaVM, LUA_GCRESTART, 0);
	
	/* Execute the Lua script */
	if (luaL_dofile(luaVM, argv[1])) err_message("Problem running script.");
	
	/* Lua cleanup */
	lua_close(luaVM);


	// detach cable (made by handle_del, too)
	err = ticalcs_cable_detach(calc);

	// remove calc & cable
	ticalcs_handle_del(calc);
	ticables_handle_del(cable);

	return 0;
}


